@@ -66,6 +66,8 @@ static struct autofs_point {
   struct lookup_mod *lookup; /* Lookup module */
 } ap;
 
+volatile struct pending_mount *junk_mounts = NULL;
+
 #define DEFAULT_TIMEOUT (5*60)	/* 5 minutes */
 #define CHECK_RATIO     4	/* exp_runfreq = exp_timeout/CHECK_RATIO */
 
@@ -331,60 +333,71 @@ static int handle_packet(void)
 {
   struct autofs_packet_missing pkt;
   struct stat st;
+  sigset_t allsignals, oldsig, chldsig;
   pid_t f;
   
   if (get_pkt(ap.pipefd, &pkt))
     return -1;
   
+  sigemptyset(&chldsig);
+  sigaddset(&chldsig, SIGCHLD);
+  sigfillset(&allsignals);
+
   chdir(ap.path);
   if ( lstat(pkt.name,&st) || (S_ISDIR(st.st_mode) && st.st_dev == ap.dev) ) {
     /* Need to mount or symlink */
     struct pending_mount *mt;
     
-    if ( !(mt = malloc(sizeof(struct pending_mount))) ) {
+    /* Block SIGCHLD while mucking with linked lists */
+    sigprocmask(SIG_BLOCK, &chldsig, NULL);
+    if ( (mt = (struct pending_mount *) junk_mounts) ) {
+      junk_mounts = junk_mounts->next;
+    } else if ( !(mt = malloc(sizeof(struct pending_mount))) ) {
       syslog(LOG_ERR, "handle_packet: malloc: %m");
       send_fail(pkt.wait_queue_token);
       return 1;
     }
-    
+    sigprocmask(SIG_UNBLOCK, &chldsig, NULL);
+
     syslog(LOG_INFO, "attempting to mount entry %s/%s",
 	   ap.path, pkt.name);
-    
+
+    sigprocmask(SIG_BLOCK, &allsignals, &oldsig);
+
     f = fork();
     if ( f == -1 ) {
+      sigprocmask(SIG_SETMASK, &oldsig, NULL);
       syslog(LOG_ERR, "handle_packet: fork: %m");
       send_fail(pkt.wait_queue_token);
       return 1;
     } else if ( !f ) {
       int i;
       
-      reset_signals();
+      reset_signals();		/* Set up a standard signal environment */
       close(ap.pipefd);
       close(ap.ioctlfd);
       
-      /* Wait for go-ahead from main process */
+      /* Wait for go-ahead from main process.  Fortunately signals sent
+	 by a process to itself are synchronous events per POSIX.1 */
       kill(getpid(), SIGSTOP);
       i = ap.lookup->lookup_mount(ap.path,pkt.name,pkt.len,ap.lookup->context);
       _exit(i ? 1 : 0);
     } else {
       int status;
-      sigset_t ss;
-      
-      sigemptyset(&ss);
-      sigaddset(&ss, SIGCHLD);
-      
+
+      sigprocmask(SIG_SETMASK, &oldsig, NULL);
       mt->pid = f;
       mt->wait_queue_token = pkt.wait_queue_token;
-      sigprocmask(SIG_BLOCK, &ss, NULL);
+      sigprocmask(SIG_BLOCK, &chldsig, NULL);
       mt->next = ap.mounts;
       ap.mounts = mt;
       if ( waitpid(f, &status, WUNTRACED) != f ||
 	   !WIFSTOPPED(status) ) {
-	sigprocmask(SIG_UNBLOCK, &ss, NULL);
+	sigprocmask(SIG_UNBLOCK, &chldsig, NULL);
 	send_fail(pkt.wait_queue_token);
 	return 1;
       }
-      sigprocmask(SIG_UNBLOCK, &ss, NULL);
+      sigprocmask(SIG_UNBLOCK, &chldsig, NULL);
       kill(f, SIGCONT); /* Go for it */
     }
   } else {
@@ -402,25 +415,27 @@ static void sig_child(int sig)
   int status;
   volatile struct pending_mount *mt, * volatile *mtp;
   
-  pid = waitpid(-1, &status, WNOHANG);
-  if ( !(pid > 0) )
-    return;
-  if ( pid == ap.exp_process ) {
-    ap.exp_process = 0;
-    return;
-  }
-  for ( mtp = &ap.mounts ; (mt = *mtp) ; mtp = &mt->next ) {
-    if ( mt->pid == pid ) {
-      if ( !WIFEXITED(status) && !WIFSIGNALED(status) )
-	break;
-      else if ( WIFSIGNALED(status) ||
-		WEXITSTATUS(status) != 0 )
-	send_fail(mt->wait_queue_token);
-      else
-	send_ready(mt->wait_queue_token);
-      *mtp = mt->next; /* Remove from linked list */
-      /*** FIXME: Need to free mt from sig handler ***/
-      break;
+  while ( (pid = waitpid(-1, &status, WNOHANG)) > 0 ) {
+    if ( pid == ap.exp_process ) {
+      ap.exp_process = 0;
+    } else {
+      for ( mtp = &ap.mounts ; (mt = *mtp) ; mtp = &mt->next ) {
+	if ( mt->pid == pid ) {
+	  if ( !WIFEXITED(status) && !WIFSIGNALED(status) )
+	    break;
+	  else if ( WIFSIGNALED(status) ||
+		    WEXITSTATUS(status) != 0 )
+	    send_fail(mt->wait_queue_token);
+	  else
+	    send_ready(mt->wait_queue_token);
+	  *mtp = mt->next; /* Remove from linked list */
+	  /* We can't use free() from a signal handler, so put this on
+	     a "junk list" which we can reclaim later */
+	  mt->next = junk_mounts;
+	  junk_mounts = mt;
+	  break;
+	}
+      }
     }
   }
 }
@@ -628,9 +643,8 @@ int main(int argc, char *argv[])
     timeout = ap.exp_timeout;
     ioctl(ap.ioctlfd, AUTOFS_IOC_SETTIMEOUT, &timeout);
     
-    /* We often start several automounters at the same
-       time.  Add some randomness so we don't all expire
-       at the same time. */
+    /* We often start several automounters at the same time.  Add some
+       randomness so we don't all expire at the same time. */
     if ( ap.exp_timeout )
       alarm(ap.exp_timeout + my_pid % ap.exp_runfreq);
   }
